(* Contexts *)
(* Author: Frank Pfenning *)
(*
   Variables (represented by name) are mapped to
   arbitrary values (normally propositions)
*)

signature CTX =
sig

  (* Contexts *)
  (* Later declarations shadow earlier ones *)
  datatype 'a Ctx =			(* G ::=       *)
    Null				(*      .      *)
  | Decl of 'a Ctx * (string * 'a)	(*    | G, x:A *)

  val lookup : 'a Ctx * string -> 'a option
  val lookupP : P.Prop Ctx * P.Prop -> string option
  val delete : P.Prop Ctx * P.Prop -> P.Prop Ctx
  val toString : P.Prop Ctx  -> string
  val add    : P.Prop Ctx * P.Prop Ctx -> P.Prop Ctx

end;  (* signature CTX *)

structure Ctx :> CTX =
struct

  (* Contexts *)
  (* Later declarations shadow earlier ones *)
  datatype 'a Ctx =			(* G ::=       *)
    Null				(*      .      *)
  | Decl of 'a Ctx * (string * 'a)	(*    | G, x:A *)

  (* lookup (G, x) = NONE if x not declared in G
                   = SOME(A) if x:A is last declaration of x in G
  *)
  fun lookup (Null, x) = NONE
    | lookup (Decl (G, (y, A)), x) =
      if (x = y) then SOME(A)
      else lookup (G, x)

  (* lookupP (G, P) = NONE if P is not in G
                    = SOME(x) if x:A is last declaration of A in G
  *)
  fun lookupP (Null, A) = NONE
    | lookupP (Decl (G, (y, A)), B) =
      if (P.eq(A, B)) then SOME(y)
      else lookupP (G, B)


  fun delete (Null, P) = Null
    | delete (Decl(G, (y,A)), P) = 
      if (P.eq (A, P)) then 
	  delete (G, P)
      else Decl(delete(G, P), (y,A))

  fun toString Null = " "
    | toString (Decl(G, (y,A))) = (toString G)^", "^(P.toString A)


  fun add (C, Null) = C
    | add (C, Decl(G, (y,A))) = Decl(add (C, G), (y,A))


end;  (* structure Ctx *)

(* Global abbreviation *)
structure C = Ctx;
